from setuptools import find_packages, setup

setup(
    name="{{ project_name }}",
    version="0.0.1",
    packages=find_packages(),
    package_data={
        "{{ project_name }}": [
            "dbt-project/**/*",
        ],
    },
    install_requires=[
        "dagster",
        "dagster-cloud",
        "dagster-dbt",
        {%- for dbt_adapter in dbt_adapter_packages %}
        "{{ dbt_adapter }}<{{ dbt_core_version_upper_bound }}",
        {%- endfor %}
    ],
    extras_require={
        "dev": [
            "dagster-webserver",
        ]
    },
)
